@@ -1,5 +1,6 @@
 # Copyright (c) Microsoft Corporation.
 # Licensed under the MIT License
+import abc
 import io
 import json
 import socket
@@ -8,18 +9,18 @@ import uuid
 from copy import copy
 from datetime import timedelta
 from enum import Enum, unique
-from typing import TYPE_CHECKING, Union, Callable, Optional, Any, Coroutine
+from typing import TYPE_CHECKING, Union, Callable, Optional, Any, Coroutine, List, Tuple, AnyStr, IO, NoReturn
 
 import requests
 from requests import Response
-from requests.adapters import HTTPAdapter
 from urllib3.connection import HTTPConnection
 
 from ._version import VERSION
 from .data_format import DataFormat
-from .exceptions import KustoServiceError
-from .response import KustoResponseDataSetV1, KustoResponseDataSetV2, KustoResponseDataSet
+from .exceptions import KustoServiceError, KustoApiError, KustoThrottlingError
+from .response import KustoResponseDataSetV1, KustoResponseDataSetV2, KustoStreamingResponseDataSet, KustoResponseDataSet
 from .security import _AadHelper
+from .streaming_response import StreamingDataSetEnumerator, JsonTokenReader
 
 if TYPE_CHECKING:
     import aiohttp
@@ -60,7 +61,7 @@ class KustoConnectionStringBuilder:
         domain_hint = "Domain Hint"
 
         @classmethod
-        def parse(cls, key: str) -> "ValidKeywords":
+        def parse(cls, key: str) -> "KustoConnectionStringBuilder.ValidKeywords":
             """Create a valid keyword."""
             key = key.lower().strip()
             if key in ["data source", "addr", "address", "network address", "server"]:
@@ -91,6 +92,14 @@ class KustoConnectionStringBuilder:
                 return cls.msi_auth
             if key in ["msi_type"]:
                 return cls.msi_params
+            if key in ["az cli"]:
+                return cls.az_cli
+            if key in ["interactive login"]:
+                return cls.interactive_login
+            if key in ["login hint"]:
+                return cls.login_hint
+            if key in ["domain hint"]:
+                return cls.domain_hint
             raise KeyError(key)
 
         def is_secret(self) -> bool:
@@ -125,7 +134,7 @@ class KustoConnectionStringBuilder:
     def __init__(self, connection_string: str):
         """
         Creates new KustoConnectionStringBuilder.
-        :param str connection_string: Kusto connection string should by of the format:
+        :param str connection_string: Kusto connection string should be of the format:
         https://<clusterName>.kusto.windows.net;AAD User ID="user@microsoft.com";Password=P@ssWord
         For more information please look at:
         https://kusto.azurewebsites.net/docs/concepts/kusto_connection_strings.html
@@ -152,7 +161,7 @@ class KustoConnectionStringBuilder:
                 else:
                     raise KeyError("Expected aad federated security to be bool. Recieved %s" % value)
 
-    def __setitem__(self, key, value):
+    def __setitem__(self, key: "Union[KustoConnectionStringBuilder.ValidKeywords, str]", value: Union[str, bool, dict]):
         try:
             keyword = key if isinstance(key, self.ValidKeywords) else self.ValidKeywords.parse(key)
         except KeyError:
@@ -181,7 +190,7 @@ class KustoConnectionStringBuilder:
         """
         Creates a KustoConnection string builder that will authenticate with AAD user name and
         password.
-        :param str connection_string: Kusto connection string should by of the format: https://<clusterName>.kusto.windows.net
+        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
         :param str user_id: AAD user ID.
         :param str password: Corresponding password of the AAD user.
         :param str authority_id: optional param. defaults to "common"
@@ -201,7 +210,7 @@ class KustoConnectionStringBuilder:
         """
         Creates a KustoConnection string builder that will authenticate with AAD application and
         a certificate credentials.
-        :param str connection_string: Kusto connection string should by of the format:
+        :param str connection_string: Kusto connection string should be of the format:
         https://<clusterName>.kusto.windows.net
         :param str user_token: AAD user token.
         """
@@ -218,7 +227,7 @@ class KustoConnectionStringBuilder:
     ) -> "KustoConnectionStringBuilder":
         """
         Creates a KustoConnection string builder that will authenticate with AAD application and key.
-        :param str connection_string: Kusto connection string should by of the format: https://<clusterName>.kusto.windows.net
+        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
         :param str aad_app_id: AAD application ID.
         :param str app_key: Corresponding key of the AAD application.
         :param str authority_id: Authority id (aka Tenant id) must be provided
@@ -241,7 +250,7 @@ class KustoConnectionStringBuilder:
         """
         Creates a KustoConnection string builder that will authenticate with AAD application using
         a certificate.
-        :param str connection_string: Kusto connection string should by of the format:
+        :param str connection_string: Kusto connection string should be of the format:
         https://<clusterName>.kusto.windows.net
         :param str aad_app_id: AAD application ID.
         :param str certificate: A PEM encoded certificate private key.
@@ -269,7 +278,7 @@ class KustoConnectionStringBuilder:
         """
         Creates a KustoConnection string builder that will authenticate with AAD application using
         a certificate Subject Name and Issuer.
-        :param str connection_string: Kusto connection string should by of the format:
+        :param str connection_string: Kusto connection string should be of the format:
         https://<clusterName>.kusto.windows.net
         :param str aad_app_id: AAD application ID.
         :param str private_certificate: A PEM encoded certificate private key.
@@ -298,7 +307,7 @@ class KustoConnectionStringBuilder:
         """
         Creates a KustoConnection string builder that will authenticate with AAD application and
         an application token.
-        :param str connection_string: Kusto connection string should by of the format:
+        :param str connection_string: Kusto connection string should be of the format:
         https://<clusterName>.kusto.windows.net
         :param str application_token: AAD application token.
         """
@@ -314,7 +323,7 @@ class KustoConnectionStringBuilder:
         """
         Creates a KustoConnection string builder that will authenticate with AAD application and
         password.
-        :param str connection_string: Kusto connection string should by of the format: https://<clusterName>.kusto.windows.net
+        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
         :param str authority_id: optional param. defaults to "common"
         """
         kcsb = cls(connection_string)
@@ -328,7 +337,7 @@ class KustoConnectionStringBuilder:
         """
         Creates a KustoConnection string builder that will use existing authenticated az cli profile
         password.
-        :param str connection_string: Kusto connection string should by of the format: https://<clusterName>.kusto.windows.net
+        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
         """
         kcsb = cls(connection_string)
         kcsb[kcsb.ValidKeywords.az_cli] = True
@@ -345,7 +354,7 @@ class KustoConnectionStringBuilder:
         an application token obtained from a Microsoft Service Identity endpoint. An optional user
         assigned application ID can be added to the token.
 
-        :param str connection_string: Kusto connection string should by of the format: https://<clusterName>.kusto.windows.net
+        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
         :param client_id: an optional user assigned identity provided as an Azure ID of a client
         :param object_id: an optional user assigned identity provided as an Azure ID of an object
         :param msi_res_id: an optional user assigned identity provided as an Azure ID of an MSI resource
@@ -366,6 +375,7 @@ class KustoConnectionStringBuilder:
         if object_id is not None:
             # Until we upgrade azure-identity to version 1.4.1, only client_id is excepted as a hint for user managed service identity
             raise ValueError("User Managed Service Identity with object_id is temporarily not supported by azure identity 1.3.1. Please use client_id instead.")
+            # noinspection PyUnreachableCode
             params["object_id"] = object_id
             exclusive_pcount += 1
 
@@ -374,6 +384,7 @@ class KustoConnectionStringBuilder:
             raise ValueError(
                 "User Managed Service Identity with msi_res_id is temporarily not supported by azure identity 1.3.1. Please use client_id instead."
             )
+            # noinspection PyUnreachableCode
             params["msi_res_id"] = msi_res_id
             exclusive_pcount += 1
 
@@ -390,7 +401,7 @@ class KustoConnectionStringBuilder:
     def with_token_provider(cls, connection_string: str, token_provider: Callable[[], str]) -> "KustoConnectionStringBuilder":
         """
         Create a KustoConnectionStringBuilder that uses a callback function to obtain a connection token
-        :param str connection_string: Kusto connection string should by of the format: https://<clusterName>.kusto.windows.net
+        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
         :param token_provider: a parameterless function that returns a valid bearer token for the relevant kusto resource as a string
         """
 
@@ -410,7 +421,7 @@ class KustoConnectionStringBuilder:
     ) -> "KustoConnectionStringBuilder":
         """
         Create a KustoConnectionStringBuilder that uses an async callback function to obtain a connection token
-        :param str connection_string: Kusto connection string should by of the format: https://<clusterName>.kusto.windows.net
+        :param str connection_string: Kusto connection string should be of the format: https://<clusterName>.kusto.windows.net
         :param async_token_provider: a parameterless function that after awaiting returns a valid bearer token for the relevant kusto resource as a string
         """
 
@@ -438,108 +449,108 @@ class KustoConnectionStringBuilder:
         return kcsb
 
     @property
-    def data_source(self) -> str:
+    def data_source(self) -> Optional[str]:
         """The URI specifying the Kusto service endpoint.
         For example, https://kuskus.kusto.windows.net or net.tcp://localhost
         """
         return self._internal_dict.get(self.ValidKeywords.data_source)
 
     @property
-    def aad_user_id(self) -> str:
+    def aad_user_id(self) -> Optional[str]:
         """The username to use for AAD Federated AuthN."""
         return self._internal_dict.get(self.ValidKeywords.aad_user_id)
 
     @property
-    def password(self) -> str:
+    def password(self) -> Optional[str]:
         """The password to use for authentication when username/password authentication is used.
         Must be accompanied by UserID property
         """
         return self._internal_dict.get(self.ValidKeywords.password)
 
     @property
-    def application_client_id(self) -> str:
+    def application_client_id(self) -> Optional[str]:
         """The application client id to use for authentication when federated
         authentication is used.
         """
         return self._internal_dict.get(self.ValidKeywords.application_client_id)
 
     @property
-    def application_key(self) -> str:
+    def application_key(self) -> Optional[str]:
         """The application key to use for authentication when federated authentication is used"""
         return self._internal_dict.get(self.ValidKeywords.application_key)
 
     @property
-    def application_certificate(self) -> str:
+    def application_certificate(self) -> Optional[str]:
         """A PEM encoded certificate private key."""
         return self._internal_dict.get(self.ValidKeywords.application_certificate)
 
     @application_certificate.setter
-    def application_certificate(self, value):
+    def application_certificate(self, value: str):
         self[self.ValidKeywords.application_certificate] = value
 
     @property
-    def application_certificate_thumbprint(self):
+    def application_certificate_thumbprint(self) -> Optional[str]:
         """hex encoded thumbprint of the certificate."""
         return self._internal_dict.get(self.ValidKeywords.application_certificate_thumbprint)
 
     @application_certificate_thumbprint.setter
-    def application_certificate_thumbprint(self, value):
+    def application_certificate_thumbprint(self, value: str):
         self[self.ValidKeywords.application_certificate_thumbprint] = value
 
     @property
-    def application_public_certificate(self) -> str:
+    def application_public_certificate(self) -> Optional[str]:
         """A public certificate matching the PEM encoded certificate private key."""
         return self._internal_dict.get(self.ValidKeywords.public_application_certificate)
 
     @application_public_certificate.setter
-    def application_public_certificate(self, value):
+    def application_public_certificate(self, value: str):
         self[self.ValidKeywords.public_application_certificate] = value
 
     @property
-    def authority_id(self):
+    def authority_id(self) -> Optional[str]:
         """The ID of the AAD tenant where the application is configured.
         (should be supplied only for non-Microsoft tenant)"""
         return self._internal_dict.get(self.ValidKeywords.authority_id)
 
     @authority_id.setter
-    def authority_id(self, value):
+    def authority_id(self, value: str):
         self[self.ValidKeywords.authority_id] = value
 
     @property
-    def aad_federated_security(self):
+    def aad_federated_security(self) -> Optional[bool]:
         """A Boolean value that instructs the client to perform AAD federated authentication."""
         return self._internal_dict.get(self.ValidKeywords.aad_federated_security)
 
     @property
-    def user_token(self):
+    def user_token(self) -> Optional[str]:
         """User token."""
         return self._internal_dict.get(self.ValidKeywords.user_token)
 
     @property
-    def application_token(self):
+    def application_token(self) -> Optional[str]:
         """Application token."""
         return self._internal_dict.get(self.ValidKeywords.application_token)
 
     @property
-    def msi_authentication(self):
+    def msi_authentication(self) -> Optional[bool]:
         """A value stating the MSI identity type to obtain"""
         return self._internal_dict.get(self.ValidKeywords.msi_auth)
 
     @property
-    def msi_parameters(self):
+    def msi_parameters(self) -> Optional[dict]:
         """A user assigned MSI ID to be obtained"""
         return self._internal_dict.get(self.ValidKeywords.msi_params)
 
     @property
-    def az_cli(self):
+    def az_cli(self) -> Optional[bool]:
         return self._internal_dict.get(self.ValidKeywords.az_cli)
 
     @property
-    def token_provider(self):
+    def token_provider(self) -> Optional[Callable[[], str]]:
         return self._token_provider
 
     @property
-    def async_token_provider(self):
+    def async_token_provider(self) -> Optional[Callable[[], Coroutine[None, None, str]]]:
         return self._async_token_provider
 
     @property
@@ -555,21 +566,21 @@ class KustoConnectionStringBuilder:
     def domain_hint(self) -> Optional[str]:
         return self._internal_dict.get(self.ValidKeywords.domain_hint)
 
-    def __str__(self):
+    def __str__(self) -> str:
         dict_copy = self._internal_dict.copy()
         for key in dict_copy:
             if key.is_secret():
                 dict_copy[key] = "****"
         return self._build_connection_string(dict_copy)
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return self._build_connection_string(self._internal_dict)
 
-    def _build_connection_string(self, kcsb_as_dict) -> str:
+    def _build_connection_string(self, kcsb_as_dict: dict) -> str:
         return ";".join(["{0}={1}".format(word.value, kcsb_as_dict[word]) for word in self.ValidKeywords if word in kcsb_as_dict])
 
 
-def _assert_value_is_valid(value):
+def _assert_value_is_valid(value: str):
     if not value or not value.strip():
         raise ValueError("Should not be empty")
 
@@ -595,34 +606,34 @@ class ClientRequestProperties:
         _assert_value_is_valid(name)
         self._parameters[name] = value
 
-    def has_parameter(self, name):
+    def has_parameter(self, name: str) -> bool:
         """Checks if a parameter is specified."""
         return name in self._parameters
 
-    def get_parameter(self, name, default_value):
+    def get_parameter(self, name: str, default_value: str) -> str:
         """Gets a parameter's value."""
         return self._parameters.get(name, default_value)
 
-    def set_option(self, name, value):
+    def set_option(self, name: str, value: Any):
         """Sets an option's value"""
         _assert_value_is_valid(name)
         self._options[name] = value
 
-    def has_option(self, name):
+    def has_option(self, name: str) -> bool:
         """Checks if an option is specified."""
         return name in self._options
 
-    def get_option(self, name, default_value):
+    def get_option(self, name: str, default_value: Any) -> str:
         """Gets an option's value."""
         return self._options.get(name, default_value)
 
-    def to_json(self):
+    def to_json(self) -> str:
         """Safe serialization to a JSON string."""
         return json.dumps({"Options": self._options, "Parameters": self._parameters}, default=str)
 
 
 class ExecuteRequestParams:
-    def __init__(self, database: str, payload: io.IOBase, properties: ClientRequestProperties, query: str, timeout: timedelta, request_headers: dict):
+    def __init__(self, database: str, payload: Optional[io.IOBase], properties: ClientRequestProperties, query: str, timeout: timedelta, request_headers: dict):
         request_headers = copy(request_headers)
         json_payload = None
         if not payload:
@@ -636,7 +647,8 @@ class ExecuteRequestParams:
             if properties:
                 request_headers.update(json.loads(properties.to_json())["Options"])
 
-            client_request_id_prefix = "KPC.execute_streaming_ingest;"
+            # Before 3.0 it was KPC.execute_streaming_ingest, but was changed to align with the other SDKs
+            client_request_id_prefix = "KPC.executeStreamingIngest;"
             request_headers["Content-Encoding"] = "gzip"
         request_headers["x-ms-client-request-id"] = client_request_id_prefix + str(uuid.uuid4())
         if properties is not None:
@@ -671,22 +683,40 @@ class HTTPAdapterWithSocketOptions(requests.adapters.HTTPAdapter):
         super(HTTPAdapterWithSocketOptions, self).init_poolmanager(*args, **kwargs)
 
 
-class _KustoClientBase:
+class _KustoClientBase(abc.ABC):
+    API_VERSION = "2019-02-13"
+
     _mgmt_default_timeout = timedelta(hours=1, seconds=30)
     _query_default_timeout = timedelta(minutes=4, seconds=30)
     _streaming_ingest_default_timeout = timedelta(minutes=10)
 
-    def __init__(self, kcsb: Union[KustoConnectionStringBuilder, str]):
+    _aad_helper: _AadHelper
+
+    def __init__(self, kcsb: Union[KustoConnectionStringBuilder, str], is_async):
         self._kcsb = kcsb
+        self._proxy_url: Optional[str] = None
         if not isinstance(kcsb, KustoConnectionStringBuilder):
             self._kcsb = KustoConnectionStringBuilder(kcsb)
         self._kusto_cluster = self._kcsb.data_source
 
+        # notice that in this context, federated actually just stands for aad auth, not aad federated auth (legacy code)
+        self._aad_helper = _AadHelper(self._kcsb, is_async) if self._kcsb.aad_federated_security else None
+
         # Create a session object for connection pooling
         self._mgmt_endpoint = "{0}/v1/rest/mgmt".format(self._kusto_cluster)
         self._query_endpoint = "{0}/v2/rest/query".format(self._kusto_cluster)
         self._streaming_ingest_endpoint = "{0}/v1/rest/ingest/".format(self._kusto_cluster)
-        self._request_headers = {"Accept": "application/json", "Accept-Encoding": "gzip,deflate", "x-ms-client-version": "Kusto.Python.Client:" + VERSION}
+        self._request_headers = {
+            "Accept": "application/json",
+            "Accept-Encoding": "gzip,deflate",
+            "x-ms-client-version": "Kusto.Python.Client:" + VERSION,
+            "x-ms-version": self.API_VERSION,
+        }
+
+    def set_proxy(self, proxy_url: str):
+        self._proxy_url = proxy_url
+        if self._aad_helper:
+            self._aad_helper.token_provider.set_proxy(proxy_url)
 
     @staticmethod
     def _kusto_parse_by_endpoint(endpoint: str, response_json: Any) -> KustoResponseDataSet:
@@ -703,21 +733,26 @@ class _KustoClientBase:
         status: int,
         response_json: Any,
         response_text: Optional[str],
-    ):  # TODO: This method return type should be "NoReturn". Re-add when the minimum python version is increased (3.6.2 or higher)
+    ) -> NoReturn:
 
         if status == 404:
             if payload:
                 raise KustoServiceError("The ingestion endpoint does not exist. Please enable streaming ingestion on your cluster.", response) from exception
 
-            raise KustoServiceError("The requested endpoint '{}' does not exist.".format(endpoint), response) from exception
+            raise KustoServiceError(f"The requested endpoint '{endpoint}' does not exist.", response) from exception
+
+        if status == 429:
+            raise KustoThrottlingError("The request was throttled by the server.", response) from exception
 
         if payload:
-            raise KustoServiceError(
-                "An error occurred while trying to ingest: Status: {0.status_code}, Reason: {0.reason}, Text: {1}.".format(response, response_text), response
-            ) from exception
+            message = f"An error occurred while trying to ingest: Status: {status}, Reason: {response.reason}, Text: {response_text}."
+            if response_json:
+                raise KustoApiError(response_json, message, response) from exception
+
+            raise KustoServiceError(message, response) from exception
 
         if response_json:
-            raise KustoServiceError([response_json], response) from exception
+            raise KustoApiError(response_json, http_response=response) from exception
 
         if response_text:
             raise KustoServiceError(response_text, response) from exception
@@ -750,20 +785,22 @@ class KustoClient(_KustoClientBase):
         :param kcsb: The connection string to initialize KustoClient.
         :type kcsb: azure.kusto.data.KustoConnectionStringBuilder or str
         """
-        super().__init__(kcsb)
+        super().__init__(kcsb, False)
 
         # Create a session object for connection pooling
         self._session = requests.Session()
+
         adapter = HTTPAdapterWithSocketOptions(
             socket_options=(HTTPConnection.default_socket_options or []) + self.compose_socket_options(), pool_maxsize=self._max_pool_size
         )
         self._session.mount("http://", adapter)
         self._session.mount("https://", adapter)
 
-        # notice that in this context, federated actually just stands for add auth, not aad federated auth (legacy code)
-        self._auth_provider = _AadHelper(self._kcsb, is_async=False) if self._kcsb.aad_federated_security else None
+    def set_proxy(self, proxy_url: str):
+        super().set_proxy(proxy_url)
+        self._session.proxies = {"http": proxy_url, "https": proxy_url}
 
-    def set_http_retries(self, max_retries):
+    def set_http_retries(self, max_retries: int):
         """
         Set the number of HTTP retries to attempt
         """
@@ -776,7 +813,7 @@ class KustoClient(_KustoClientBase):
         self._session.mount("https://", adapter)
 
     @staticmethod
-    def compose_socket_options():
+    def compose_socket_options() -> List[Tuple[int, int, int]]:
         # Sends TCP Keep-Alive after MAX_IDLE_SECONDS seconds of idleness, once every INTERVAL_SECONDS seconds, and closes the connection after MAX_FAILED_KEEPALIVES failed pings (e.g. 20 => 1:00:30)
         MAX_IDLE_SECONDS = 30
         INTERVAL_SECONDS = 180  # Corresponds to Azure Load Balancer Service 4 minute timeout, with 1 minute of slack
@@ -814,7 +851,7 @@ class KustoClient(_KustoClientBase):
         else:
             return []
 
-    def execute(self, database: str, query: str, properties: ClientRequestProperties = None) -> KustoResponseDataSet:
+    def execute(self, database: str, query: str, properties: Optional[ClientRequestProperties] = None) -> KustoResponseDataSet:
         """
         Executes a query or management command.
         :param str database: Database against query will be executed.
@@ -828,7 +865,7 @@ class KustoClient(_KustoClientBase):
             return self.execute_mgmt(database, query, properties)
         return self.execute_query(database, query, properties)
 
-    def execute_query(self, database: str, query: str, properties: ClientRequestProperties = None) -> KustoResponseDataSet:
+    def execute_query(self, database: str, query: str, properties: Optional[ClientRequestProperties] = None) -> KustoResponseDataSet:
         """
         Execute a KQL query.
         To learn more about KQL go to https://docs.microsoft.com/en-us/azure/kusto/query/
@@ -840,7 +877,7 @@ class KustoClient(_KustoClientBase):
         """
         return self._execute(self._query_endpoint, database, query, None, self._query_default_timeout, properties)
 
-    def execute_mgmt(self, database: str, query: str, properties: ClientRequestProperties = None) -> KustoResponseDataSet:
+    def execute_mgmt(self, database: str, query: str, properties: Optional[ClientRequestProperties] = None) -> KustoResponseDataSet:
         """
         Execute a KQL control command.
         To learn more about KQL control commands go to  https://docs.microsoft.com/en-us/azure/kusto/management/
@@ -856,9 +893,9 @@ class KustoClient(_KustoClientBase):
         self,
         database: str,
         table: str,
-        stream: io.IOBase,
+        stream: IO[AnyStr],
         stream_format: Union[DataFormat, str],
-        properties: ClientRequestProperties = None,
+        properties: Optional[ClientRequestProperties] = None,
         mapping_name: str = None,
     ):
         """
@@ -873,32 +910,66 @@ class KustoClient(_KustoClientBase):
         :param ClientRequestProperties properties: additional request properties.
         :param str mapping_name: Pre-defined mapping of the table. Required when stream_format is json/avro.
         """
-        stream_format = stream_format.value if isinstance(stream_format, DataFormat) else DataFormat(stream_format.lower()).value
+        stream_format = stream_format.kusto_value if isinstance(stream_format, DataFormat) else DataFormat[stream_format.upper()].kusto_value
         endpoint = self._streaming_ingest_endpoint + database + "/" + table + "?streamFormat=" + stream_format
         if mapping_name is not None:
             endpoint = endpoint + "&mappingName=" + mapping_name
 
         self._execute(endpoint, database, None, stream, self._streaming_ingest_default_timeout, properties)
 
+    def _execute_streaming_query_parsed(
+        self, database: str, query: str, timeout: timedelta = _KustoClientBase._query_default_timeout, properties: Optional[ClientRequestProperties] = None
+    ) -> StreamingDataSetEnumerator:
+        response = self._execute(self._query_endpoint, database, query, None, timeout, properties, stream_response=True)
+        response.raw.decode_content = True
+        return StreamingDataSetEnumerator(JsonTokenReader(response.raw))
+
+    def execute_streaming_query(
+        self, database: str, query: str, timeout: timedelta = _KustoClientBase._query_default_timeout, properties: Optional[ClientRequestProperties] = None
+    ) -> KustoStreamingResponseDataSet:
+        """
+        Execute a KQL query without reading it all to memory.
+        The resulting KustoStreamingResponseDataSet will stream one table at a time, and the rows can be retrieved sequentially.
+
+        :param str database: Database against query will be executed.
+        :param str query: Query to be executed.
+        :param timedelta timeout: timeout for the query to be executed
+        :param azure.kusto.data.ClientRequestProperties properties: Optional additional properties.
+        :return KustoStreamingResponseDataSet:
+        """
+        return KustoStreamingResponseDataSet(self._execute_streaming_query_parsed(database, query, timeout, properties))
+
     def _execute(
-        self, endpoint: str, database: str, query: Optional[str], payload: Optional[io.IOBase], timeout: timedelta, properties: ClientRequestProperties = None
-    ):
+        self,
+        endpoint: str,
+        database: str,
+        query: Optional[str],
+        payload: Optional[IO[AnyStr]],
+        timeout: timedelta,
+        properties: Optional[ClientRequestProperties] = None,
+        stream_response: bool = False,
+    ) -> Union[KustoResponseDataSet, Response]:
         """Executes given query against this client"""
         request_params = ExecuteRequestParams(database, payload, properties, query, timeout, self._request_headers)
         json_payload = request_params.json_payload
         request_headers = request_params.request_headers
         timeout = request_params.timeout
+        if self._aad_helper:
+            request_headers["Authorization"] = self._aad_helper.acquire_authorization_header()
+        response = self._session.post(endpoint, headers=request_headers, json=json_payload, data=payload, timeout=timeout.seconds, stream=stream_response)
 
-        if self._auth_provider:
-            request_headers["Authorization"] = self._auth_provider.acquire_authorization_header()
-
-        response = self._session.post(endpoint, headers=request_headers, data=payload, json=json_payload, timeout=timeout.seconds)
+        if stream_response:
+            try:
+                response.raise_for_status()
+                return response
+            except Exception as e:
+                raise self._handle_http_error(e, self._query_endpoint, None, response, response.status_code, response.json(), response.text)
 
         response_json = None
         try:
             response_json = response.json()
             response.raise_for_status()
         except Exception as e:
-            self._handle_http_error(e, endpoint, payload, response, response.status_code, response_json, response.text)
+            raise self._handle_http_error(e, endpoint, payload, response, response.status_code, response_json, response.text)
 
         return self._kusto_parse_by_endpoint(endpoint, response_json)
